/*
This source file is part of ProceduralGenerator (https://sourceforge.net/projects/proceduralgene/)

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along with
the program; if not, write to the Free Software Foundation, Inc., 59 Temple
Place - Suite 330, Boston, MA 02111-1307, USA, or go to
http://www.gnu.org/copyleft/lesser.txt.
*/
#ifndef ___SimiliLifeGame___
#define ___SimiliLifeGame___

#include <LibGenerator/ProceduralGenerator.h>

namespace ProceduralTextures
{
	class BlgCell
	{
	private:
		UbPixel * m_alivePx;
		UbPixel * m_deadPx;

	public:
		BlgCell * m_neighbours[8];
		bool m_alive;
		UbPixel * m_nextPixel;
		UiPixel m_pixel;
		UiPixel m_step;
		UiPixel m_age;
		UiPixel m_med;

	public:
		BlgCell();
		BlgCell( const BlgCell & cell);
		~BlgCell();

		void Set( UbPixel * p_pixel, UbPixel * p_alivePx, UbPixel * p_deadPx, bool p_alive);
		void Die();
		void Live();
		void Stay();

		void SetAlive( bool p_alive);

		void Update();

		BlgCell & operator =( const BlgCell & p_cell);
		BlgCell & Copy( const BlgCell & p_cell);

	private:
		void _addToPixelRed( unsigned char p_red);
		void _addToPixelGreen( unsigned char p_green);
		void _addToPixelBlue( unsigned char p_blue);
	};
}

namespace ProceduralTextures
{
	template <> class Policy<ProceduralTextures::BlgCell>
	{
	private:
		typedef ProceduralTextures::BlgCell value_type;

	public:
		static value_type	zero			()							{ return value_type(); }
		static value_type	unit			()							{ return value_type(); }
		static void			init			( value_type & p_a)			{ p_a = zero(); }
		static bool			is_null			( const value_type & p_a)	{ return equals( p_a, zero()); }
		static value_type	negate			( const value_type & p_a)	{ return p_a; }
		static value_type	ass_negate		( value_type & p_a)			{ return assign( p_a, negate( p_a)); }
		static value_type	stick			( value_type & p_a)			{ return p_a; }

		template <typename Ty> static value_type	convert			( const Ty & p_value)						{ return static_cast<value_type>( p_value); }
		template <typename Ty> static bool			equals			( const value_type & p_a, const Ty & p_b)	{ return false; }
		template <typename Ty> static value_type	add				( const value_type & p_a, const Ty & p_b)	{ return p_a; }
		template <typename Ty> static value_type	substract		( const value_type & p_a, const Ty & p_b)	{ return p_a; }
		template <typename Ty> static value_type	multiply		( const value_type & p_a, const Ty & p_b)	{ return p_a; }
		template <typename Ty> static value_type	divide			( const value_type & p_a, const Ty & p_b)	{ return p_a; }
		template <typename Ty> static value_type	ass_add			( value_type & p_a, const Ty & p_b)			{ return p_a; }
		template <typename Ty> static value_type	ass_substract	( value_type & p_a, const Ty & p_b)			{ return p_a; }
		template <typename Ty> static value_type	ass_multiply	( value_type & p_a, const Ty & p_b)			{ return p_a; }
		template <typename Ty> static value_type	ass_divide		( value_type & p_a, const Ty & p_b)			{ return p_a; }
		template <typename Ty> static value_type	assign			( value_type & p_a, const Ty & p_b)			{ return p_a = convert<Ty>( p_b); }
	};
}

namespace ProceduralTextures
{
	typedef Buffer<BlgCell> BlgCellBuffer;

	class BuggedLifeGame : public ProceduralGenerator
	{
	private:
		class Thread : public ProceduralGenerator::Thread
		{
		private:
			BlgCellBuffer * m_cells;

		public:
			Thread( ProceduralGenerator * p_pParent, size_t p_uiIndex, int iWidth, int iTop, int iBottom, int iTotalHeight, const UbPixel & p_pxColour);
			virtual ~Thread();

			virtual void Step();

			inline void SetBuffer( BlgCellBuffer * p_buffer) { m_cells = p_buffer; }

		private:
			int _getAliveNeightbours( int p_x, int p_y);
		};

	private:
		UbPixel m_alivePx;
		UbPixel m_deadPx;
		BlgCellBuffer m_cells;
		BlgCellBuffer m_oldCells;
		Specifics< eCONTROL_TYPE_STATIC > m_specStatR;
		Specifics< eCONTROL_TYPE_STATIC > m_specStatG;
		Specifics< eCONTROL_TYPE_STATIC > m_specStatB;
		Specifics< eCONTROL_TYPE_SLIDER > m_specSlideR;
		Specifics< eCONTROL_TYPE_SLIDER > m_specSlideG;
		Specifics< eCONTROL_TYPE_SLIDER > m_specSlideB;

	public:
		BuggedLifeGame( int p_width, int p_height, int iWndId, wxFrame * p_pFrame);
		virtual ~BuggedLifeGame();

		virtual void 			InitialiseStep		();

		virtual void SetRed		( unsigned char val) { m_pxColour.r = val;m_alivePx.r = val; }
		virtual void SetGreen	( unsigned char val) { m_pxColour.g = val;m_alivePx.g = val; }
		virtual void SetBlue	( unsigned char val) { m_pxColour.b = val;m_alivePx.b = val; }

	private:
		virtual void DoGlInit			(){}
		virtual void DoGlPreRender		(){}
		virtual void DoGlRender			( bool & p_bChanged){}
		virtual void DoGlPostRender		(){}
		virtual void DoGlCleanup			(){}
		virtual void DoGeneratePanel	();

		typedef enum
		{
			eRedValue	= 50,
			eGreenValue	= 51,
			eBlueValue	= 52,
		}
		eIDs;

		void OnSliderRed	( wxCommandEvent & p_event );
		void OnSliderGreen	( wxCommandEvent & p_event );
		void OnSliderBlue	( wxCommandEvent & p_event );

		void DoInitialiseNeighbours( int p_x, int p_y);
		int _getAliveNeightbours( int p_x, int p_y);
	};

	class BlgPlugin : public Plugin< BuggedLifeGame, BlgPlugin >
	{
	public:
		BlgPlugin();
		virtual ~BlgPlugin();

		virtual wxString GetName();

		DECLARE_PLUGIN()
	};
}

#endif
