#include "LibGenerator/PrecompiledHeader.h"

#include "LibGenerator/ProceduralGenerator.h"
#include "LibGenerator/GlObjects.h"

#ifdef __WXMSW__
#	include<wx/msw/msvcrt.h>
#endif

using namespace ProceduralTextures;

//*************************************************************************************************

#ifdef _WIN32
#	define LibError() GetLastError()
#else
#	define LibError() dlerror()
#endif

//*************************************************************************************************

ProceduralGenerator::Thread :: Thread( ProceduralGenerator * p_pParent, size_t p_uiIndex, int iWidth, int iTop, int iBottom, int iTotalHeight, const UbPixel & p_pxColour)
	:	wxThread	( wxTHREAD_DETACHED	)
	,	m_pParent	( p_pParent			)
	,	m_uiIndex	( p_uiIndex			)
	,	m_iWidth	( iWidth			)
	,	m_iBottom	( iBottom			)
	,	m_iTop		( iTop				)
	,	m_iHeight	( iTotalHeight		)
	,	m_pxColour	( p_pxColour		)
	,	m_bStopped	( false				)
	,	m_bLaunched	( false				)
{
}

ProceduralGenerator::Thread :: ~Thread()
{
}

wxThread::ExitCode ProceduralGenerator::Thread :: Entry()
{
	wxCommandEvent l_event( wxEVT_NULL, eEVENT_WORKER_THREAD_END );
	l_event.SetInt( m_uiIndex );

	while( ! IsStopped() )
	{
		if( IsLaunched() )
		{
			SetLaunched( false );
			// Lock thread data
			m_mutex.Lock();
			// Make computations
			Step();
			// Unlock thread data
			m_mutex.Unlock();
			// Tell parent I am ended
			m_pParent->AddPendingEvent( l_event );
		}

		Sleep( 1 );
	}

	return 0;
}

//*************************************************************************************************

ProceduralGenerator :: ProceduralGenerator( int p_width, int p_height, int p_iWndId, wxFrame * p_pMainFrame, bool p_bNeedsConfigPanel, int p_iSleepTime )
	:	wxThread				( wxTHREAD_DETACHED						)
	,	m_sizeDisplay			( p_width, p_height						)
	,	m_sizeImage				( p_width, p_height						)
	,	m_step					( 0										)
	,	m_initialised			( false									)
	,	m_pxColour				( 255, 255, 255, 255					)
	,	m_bStopped				( false									)
	,	m_frontBuffer			( Point<size_t, 2>( p_width, p_height)	)
	,	m_backBuffer			( Point<size_t, 2>( p_width, p_height)	)
	,	m_pMainFrame			( p_pMainFrame 							)
	,	m_bNeedsConfigPanel		( p_bNeedsConfigPanel					)
	,	m_bToSave				( false									)
	,	m_bRendered				( true									)
	,	m_iSleepTime			( p_iSleepTime							)
	,	m_pfnCpuStepEndHandler	( NULL									)
{
	m_frameBuffer.Resize( m_sizeImage);
	m_texture.Resize( m_sizeImage);
	m_frameBuffer.BindTexture( GL_COLOR_ATTACHMENT0, & m_texture);
	m_frameBufferSave.BindTexture( GL_COLOR_ATTACHMENT0, & m_textureSave);
	m_uiThreadCount = GetCPUCount() * 2;// - 1;
	Connect( eEVENT_WORKER_THREAD_END,	wxEVT_NULL, wxCommandEventHandler( ProceduralGenerator::OnThreadEnd	) );
	Connect( eEVENT_THREAD_STOP,		wxEVT_NULL, wxCommandEventHandler( ProceduralGenerator::OnStop 		) );
}

ProceduralGenerator :: ~ProceduralGenerator()
{
}

void ProceduralGenerator :: WaitForStepEnd()
{
	std::size_t l_uiThreadCount = m_arraySlaveThreads.size();

	while( m_uiThreadEndedCount < l_uiThreadCount && ! IsStopped() )
	{
		Sleep( 1 );
	}
}

void ProceduralGenerator :: OnThreadEnd( wxCommandEvent & p_event )
{
	int iThreadId = p_event.GetInt();
	m_uiThreadEndedCount++;
}

void ProceduralGenerator :: OnStop( wxCommandEvent & p_event )
{
	m_bStopped = true;

	if( m_pfnCpuStepEndHandler )
	{
		m_pMainFrame->GetEventHandler()->Disconnect( eEVENT_CPU_STEP_END, wxEVT_NULL );
	}
}

wxThread::ExitCode ProceduralGenerator :: Entry()
{
	ExitCode l_eReturn = 0;
	wxCommandEvent l_event( wxEVT_NULL, eEVENT_CPU_STEP_END );

	if( m_pfnCpuStepEndHandler )
	{
		m_pMainFrame->GetEventHandler()->Connect( eEVENT_CPU_STEP_END, wxEVT_NULL, m_pfnCpuStepEndHandler );
	}

	while( ! IsStopped() )
	{
		if( m_initialised && IsRendered() )
		{
			SetRendered( false );
			SwapBuffers();
			InitialiseStep();
			WaitForStepEnd();
			m_pMainFrame->GetEventHandler()->AddPendingEvent( l_event );
		}

		Sleep( m_iSleepTime );
	}

	DoCleanup();

	return l_eReturn;
}

void ProceduralGenerator :: SetRed( unsigned char val)
{
	m_pxColour.r = val;

	for (size_t i = 0 ; i < m_arraySlaveThreads.size() ; i++)
	{
		if (m_arraySlaveThreads[i] != NULL)
		{
			m_arraySlaveThreads[i]->SetRed( val);
		}
	}
}

void ProceduralGenerator :: SetGreen( unsigned char val)
{
	m_pxColour.g = val;

	for (size_t i = 0 ; i < m_arraySlaveThreads.size() ; i++)
	{
		if (m_arraySlaveThreads[i] != NULL)
		{
			m_arraySlaveThreads[i]->SetGreen( val);
		}
	}
}

void ProceduralGenerator :: SetBlue( unsigned char val)
{
	m_pxColour.b = val;

	for (size_t i = 0 ; i < m_arraySlaveThreads.size() ; i++)
	{
		if (m_arraySlaveThreads[i] != NULL)
		{
			m_arraySlaveThreads[i]->SetBlue( val);
		}
	}
}

void ProceduralGenerator :: SwapBuffers()
{
	m_frontBuffer = m_backBuffer;
}

void ProceduralGenerator :: InitialiseStep()
{
	m_uiThreadEndedCount = 0;

	for (size_t i = 0 ; i < m_arraySlaveThreads.size() ; i++)
	{
		if (m_arraySlaveThreads[i] != NULL)
		{
			m_arraySlaveThreads[i]->SetLaunched( true );
		}
	}
}

void ProceduralGenerator :: ClearAllThreads()
{
	m_initialised = false;

	for (size_t i = 0 ; i < m_arraySlaveThreads.size() ; i++)
	{
		ProceduralGenerator::Thread * l_pThread = m_arraySlaveThreads[i];
		m_arraySlaveThreads[i] = NULL;

		if( l_pThread )
		{
			l_pThread->Stop();
		}
	}

	m_arraySlaveThreads.clear();
}

void ProceduralGenerator :: SetSize( int p_iWidth, int p_iHeight)
{
	DoResize( wxSize( p_iWidth, p_iHeight));
}

void ProceduralGenerator :: SetSize( const wxSize & p_size)
{
	DoResize( p_size);
}

void ProceduralGenerator :: Resize( const wxSize & p_size)
{
	m_sizeDisplay = p_size;
}

void ProceduralGenerator :: GlInit()
{
	OpenGl::Initialise();
	OpenGl::ClearColor( 1.0, 1.0, 1.0, 1.0);
	OpenGl::ShadeModel( GL_FLAT);

	m_vertexBuffer.Initialise();
	m_indexBuffer.Initialise();
	m_texture.Initialise();
	m_frameBuffer.Initialise();
	m_frameBufferSave.Initialise();
	m_textureSave.Initialise();

	DoGlInit();
}

void ProceduralGenerator :: GlPreRender()
{
	m_texture.Initialise( & m_frontBuffer);
	m_frameBuffer.Initialise( this);
	m_textureSave.Initialise( m_sizeImage);
	m_frameBufferSave.Initialise( this);
	DoGlPreRender();
}

void ProceduralGenerator :: GlRender( bool p_bChanged)
{
	bool l_bChanged( p_bChanged);
	DoGlRender( l_bChanged);

	if (m_bToSave)
	{
		OpenGl::Viewport( 0, 0, m_sizeImage.x, m_sizeImage.y);
	}
	else
	{
		OpenGl::Viewport( 0, 0, m_sizeDisplay.x, m_sizeDisplay.y);
	}

	if (m_bToSave)
	{
		m_frameBufferSave.Activate();
	}

	OpenGl::MatrixMode( GL_PROJECTION);
	OpenGl::LoadIdentity();
	OpenGl::Ortho( 0.0, 1.0, 0.0, 1.0, 0.0, 1.0);
	OpenGl::MatrixMode( GL_MODELVIEW);
	OpenGl::LoadIdentity();
	OpenGl::Clear( GL_COLOR_BUFFER_BIT);
	m_texture.Activate();

	if (l_bChanged)
	{
		m_texture.UploadGlImageSync();
	}

	DoSubRender();
	m_texture.Deactivate();

	if (m_bToSave)
	{
		DoSaveFrame();
		m_frameBufferSave.Deactivate();
	}
}

void ProceduralGenerator :: GlPostRender()
{
	DoGlPostRender();
	SetRendered( true );
}

void ProceduralGenerator :: GlCleanup()
{
	m_vertexBuffer.Cleanup();
	m_indexBuffer.Cleanup();
	m_texture.Cleanup();
	m_frameBuffer.Cleanup();
	m_frameBufferSave.Cleanup();
	m_textureSave.Cleanup();
	DoGlCleanup();
	OpenGl::Cleanup();
}

void ProceduralGenerator :: SaveFrame()
{
	m_bToSave = true;
}

void ProceduralGenerator :: DoCleanup()
{
	ClearAllThreads();
	m_frontBuffer.clear();
	m_backBuffer.clear();
}

void ProceduralGenerator :: DoResize( const wxSize & p_size)
{
	m_sizeImage = p_size;
	m_sizeDisplay = p_size;
	wxSize l_sizeConfig( 0, 0);

	if( m_bNeedsConfigPanel )
	{
		l_sizeConfig = wxSize( 150, 0 );
	}

	wxDisplay l_display;
	wxSize l_maxSize = l_display.GetClientArea().GetSize() + m_pMainFrame->GetClientSize() - m_pMainFrame->GetSize() - wxSize( l_sizeConfig.x, 0);
	double l_dRatio = double( m_sizeDisplay.x) / double( m_sizeDisplay.y);
	bool l_bChanged = false;

	l_maxSize = wxSize( std::min<int>( l_maxSize.x, l_maxSize.y * l_dRatio),
		std::min<int>( l_maxSize.y, l_maxSize.x / l_dRatio));

	while (m_sizeDisplay.x > l_maxSize.x || m_sizeDisplay.y > l_maxSize.y)
	{
		if (m_sizeDisplay.x > l_maxSize.x)
		{
			m_sizeDisplay.x = l_maxSize.x;
			m_sizeDisplay.y = m_sizeDisplay.x / l_dRatio;
		}
		else if (m_sizeDisplay.y > l_maxSize.y)
		{
			m_sizeDisplay.y = l_maxSize.y;
			m_sizeDisplay.x = m_sizeDisplay.y * l_dRatio;
		}

		l_bChanged = true;
	}

	m_pMainFrame->SetClientSize( m_sizeDisplay + wxSize( l_sizeConfig.x, 0));
	m_texture.Resize( m_sizeImage);
	m_frameBuffer.Resize( m_sizeImage);
}

wxSize ProceduralGenerator :: DoLoadImage( const wxString & p_strImagePath, wxImage & p_image, PixelBuffer & p_buffer)
{
	wxSize l_sizeReturn;

	if (p_image.LoadFile( p_strImagePath))
	{
		wxDisplay l_display;
		wxSize l_sizeConfig;

		if( m_bNeedsConfigPanel )
		{
			l_sizeConfig = wxSize( 150, 0 );
		}

		l_sizeReturn.x = p_image.GetWidth();
		l_sizeReturn.y = p_image.GetHeight();

		p_image = p_image.Mirror( false);
		p_buffer.init( Point<size_t, 2>( l_sizeReturn.x, l_sizeReturn.y));

		if (p_image.HasAlpha())
		{
			p_buffer.set<unsigned char, 3, 1, false, false>( p_image.GetData(), p_image.GetAlpha());
		}
		else
		{
			p_buffer.set<unsigned char, 3, 0, false, false>( p_image.GetData(), NULL);
		}
	}
	else
	{
		wxMessageBox( wxT( "No handler found for this image format"));
	}

	return l_sizeReturn;
}

void ProceduralGenerator :: DoSubRender()
{
	if (m_vertexBuffer.Activate())
	{
		if (m_indexBuffer.Activate())
		{
			OpenGl::DrawElements( GL_TRIANGLES, m_indexBuffer.GetSize(), GL_UNSIGNED_INT, 0);
			m_indexBuffer.Deactivate();
		}

		m_vertexBuffer.Deactivate();
	}
}

void ProceduralGenerator :: DoSaveFrame()
{
	if (m_bToSave)
	{
		wxImage l_image( m_sizeImage.x, m_sizeImage.y, true);
		int x = 0;
		int y;
		m_frameBufferSave.DownloadBuffer( GL_COLOR_ATTACHMENT0, m_frontBuffer);

		for (int i = 0 ; i < m_sizeImage.x ; i++)
		{
			y = m_sizeImage.y - 1;
			for (int j = 0 ; j < m_sizeImage.y ; j++)
			{
				const UbPixel & l_pixel = m_frontBuffer[j][i];
				l_image.SetRGB( x, y, l_pixel.r, l_pixel.g, l_pixel.b);
				y--;
			}
			x++;
		}

		l_image.SaveFile( wxT( "./Image.bmp"), wxBITMAP_TYPE_BMP);
		m_bToSave = false;
	}
}

//*************************************************************************************************
